@@ -6,11 +6,14 @@
 
 #include "src/common/globals.h"
 #include "src/execution/protectors.h"
+#include "src/flags/flags.h"
 #include "src/objects/property-cell.h"
 
 namespace v8 {
 namespace internal {
 
+#include "src/codegen/define-code-stub-assembler-macros.inc"
+
 namespace {
 
 inline bool IsBigInt64OpSupported(BinaryOpAssembler* assembler, Operation op) {
@@ -46,6 +49,13 @@ TNode<Object> BinaryOpAssembler::Generate_AddWithFeedback(
                        rhs_known_smi ? Label::kDeferred : Label::kNonDeferred);
   Branch(TaggedIsNotSmi(lhs), &if_lhsisnotsmi, &if_lhsissmi);
 
+  auto IsAdditiveSafeIntegerFeedback = [&](TNode<Float64T> value) {
+    return Select<BoolT>(
+        IsAdditiveSafeIntegerFeedbackEnabled(),
+        [&] { return IsAdditiveSafeInteger(value); },
+        [&] { return BoolConstant(false); });
+  };
+
   BIND(&if_lhsissmi);
   {
     Comment("lhs is Smi");
@@ -63,6 +73,10 @@ TNode<Object> BinaryOpAssembler::Generate_AddWithFeedback(
 
         var_fadd_lhs = SmiToFloat64(lhs_smi);
         var_fadd_rhs = LoadHeapNumberValue(rhs_heap_object);
+        var_type_feedback = SelectSmiConstant(
+            IsAdditiveSafeIntegerFeedback(var_fadd_rhs.value()),
+            BinaryOperationFeedback::kAdditiveSafeInteger,
+            BinaryOperationFeedback::kNumber);
         Goto(&do_fadd);
       }
 
@@ -92,6 +106,10 @@ TNode<Object> BinaryOpAssembler::Generate_AddWithFeedback(
       {
         var_fadd_lhs = SmiToFloat64(lhs_smi);
         var_fadd_rhs = SmiToFloat64(rhs_smi);
+        var_type_feedback =
+            SelectSmiConstant(IsAdditiveSafeIntegerFeedbackEnabled(),
+                              BinaryOperationFeedback::kAdditiveSafeInteger,
+                              BinaryOperationFeedback::kNumber);
         Goto(&do_fadd);
       }
     }
@@ -116,6 +134,13 @@ TNode<Object> BinaryOpAssembler::Generate_AddWithFeedback(
 
         var_fadd_lhs = LoadHeapNumberValue(lhs_heap_object);
         var_fadd_rhs = LoadHeapNumberValue(rhs_heap_object);
+        var_type_feedback = SmiConstant(BinaryOperationFeedback::kNumber);
+        GotoIfNot(IsAdditiveSafeIntegerFeedback(var_fadd_lhs.value()),
+                  &do_fadd);
+        GotoIfNot(IsAdditiveSafeIntegerFeedback(var_fadd_rhs.value()),
+                  &do_fadd);
+        var_type_feedback =
+            SmiConstant(BinaryOperationFeedback::kAdditiveSafeInteger);
         Goto(&do_fadd);
       }
 
@@ -124,19 +149,38 @@ TNode<Object> BinaryOpAssembler::Generate_AddWithFeedback(
     {
       var_fadd_lhs = LoadHeapNumberValue(lhs_heap_object);
       var_fadd_rhs = SmiToFloat64(CAST(rhs));
+      var_type_feedback =
+          SelectSmiConstant(IsAdditiveSafeIntegerFeedback(var_fadd_lhs.value()),
+                            BinaryOperationFeedback::kAdditiveSafeInteger,
+                            BinaryOperationFeedback::kNumber);
       Goto(&do_fadd);
     }
   }
 
   BIND(&do_fadd);
   {
-    var_type_feedback = SmiConstant(BinaryOperationFeedback::kNumber);
-    UpdateFeedback(var_type_feedback.value(), maybe_feedback_vector(), slot_id,
-                   update_feedback_mode);
     TNode<Float64T> value =
         Float64Add(var_fadd_lhs.value(), var_fadd_rhs.value());
     TNode<HeapNumber> result = AllocateHeapNumberWithValue(value);
     var_result = result;
+
+    Label AdditiveSafeInteger_overflow_check_done(this);
+    GotoIfNot(IsAdditiveSafeIntegerFeedbackEnabled(),
+              &AdditiveSafeInteger_overflow_check_done);
+    {
+      GotoIfNot(
+          SmiEqual(var_type_feedback.value(),
+                   SmiConstant(BinaryOperationFeedback::kAdditiveSafeInteger)),
+          &AdditiveSafeInteger_overflow_check_done);
+      GotoIf(IsAdditiveSafeIntegerFeedback(value),
+             &AdditiveSafeInteger_overflow_check_done);
+      var_type_feedback = SmiConstant(BinaryOperationFeedback::kNumber);
+      Goto(&AdditiveSafeInteger_overflow_check_done);
+    }
+    BIND(&AdditiveSafeInteger_overflow_check_done);
+
+    UpdateFeedback(var_type_feedback.value(), maybe_feedback_vector(), slot_id,
+                   update_feedback_mode);
     Goto(&end);
   }
 
@@ -823,7 +867,7 @@ TNode<Object> BinaryOpAssembler::Generate_BitwiseBinaryOpWithOptionalFeedback(
   Label if_left_bigint(this), if_left_bigint64(this);
   Label if_left_number_right_bigint(this, Label::kDeferred);
 
-  FeedbackValues feedback =
+  FeedbackValues feedback_values =
       slot ? FeedbackValues{&var_left_feedback, maybe_feedback_vector, slot,
                             update_feedback_mode}
            : FeedbackValues();
@@ -831,13 +875,13 @@ TNode<Object> BinaryOpAssembler::Generate_BitwiseBinaryOpWithOptionalFeedback(
   TaggedToWord32OrBigIntWithFeedback(
       context(), left, &if_left_number, &var_left_word32, &if_left_bigint,
       IsBigInt64OpSupported(this, bitwise_op) ? &if_left_bigint64 : nullptr,
-      &var_left_bigint, feedback);
+      &var_left_bigint, feedback_values);
 
   BIND(&if_left_number);
-  feedback.var_feedback = slot ? &var_right_feedback : nullptr;
+  feedback_values.var_feedback = slot ? &var_right_feedback : nullptr;
   TaggedToWord32OrBigIntWithFeedback(
       context(), right, &do_number_op, &var_right_word32,
-      &if_left_number_right_bigint, nullptr, nullptr, feedback);
+      &if_left_number_right_bigint, nullptr, nullptr, feedback_values);
 
   BIND(&if_left_number_right_bigint);
   {
@@ -1118,5 +1162,7 @@ BinaryOpAssembler::Generate_BitwiseBinaryOpWithSmiOperandAndOptionalFeedback(
   return result.value();
 }
 
+#include "src/codegen/undef-code-stub-assembler-macros.inc"
+
 }  // namespace internal
 }  // namespace v8
