@@ -907,7 +907,7 @@ const fatalExceptionStackEnhancers = {
     // ANSI escape sequences is not reliable.
     if (isWindows) {
       const info = internalBinding('os').getOSInformation();
-      const ver = ArrayPrototypeMap(StringPrototypeSplit(info[2], '.'),
+      const ver = ArrayPrototypeMap(StringPrototypeSplit(info[2], '.', 3),
                                     Number);
       if (ver[0] !== 10 || ver[2] < 14393) {
         useColors = false;
@@ -1219,6 +1219,7 @@ E('ERR_FS_CP_SYMLINK_TO_SUBDIRECTORY',
 E('ERR_FS_CP_UNKNOWN', 'Cannot copy an unknown file type', SystemError);
 E('ERR_FS_EISDIR', 'Path is a directory', SystemError, HideStackFramesError);
 E('ERR_FS_FILE_TOO_LARGE', 'File size (%s) is greater than 2 GiB', RangeError);
+E('ERR_FS_WATCH_QUEUE_OVERFLOW', 'fs.watch() queued more than %d events', Error);
 E('ERR_HTTP2_ALTSVC_INVALID_ORIGIN',
   'HTTP/2 ALTSVC frames require a valid origin', TypeError);
 E('ERR_HTTP2_ALTSVC_LENGTH',
@@ -1338,7 +1339,12 @@ E('ERR_IMPORT_ATTRIBUTE_MISSING',
 E('ERR_IMPORT_ATTRIBUTE_TYPE_INCOMPATIBLE',
   'Module "%s" is not of type "%s"', TypeError);
 E('ERR_IMPORT_ATTRIBUTE_UNSUPPORTED',
-  'Import attribute "%s" with value "%s" is not supported', TypeError);
+  function error(attribute, value, url = undefined) {
+    if (url === undefined) {
+      return `Import attribute "${attribute}" with value "${value}" is not supported`;
+    }
+    return `Import attribute "${attribute}" with value "${value}" is not supported in ${url}`;
+  }, TypeError);
 E('ERR_INCOMPATIBLE_OPTION_PAIR',
   'Option "%s" cannot be used in combination with option "%s"', TypeError, HideStackFramesError);
 E('ERR_INPUT_TYPE_NOT_ALLOWED', '--input-type can only be used with string ' +
@@ -1522,7 +1528,7 @@ E('ERR_INVALID_SYNC_FORK_INPUT',
   'Asynchronous forks do not support ' +
     'Buffer, TypedArray, DataView or string input: %s',
   TypeError);
-E('ERR_INVALID_THIS', 'Value of "this" must be of type %s', TypeError);
+E('ERR_INVALID_THIS', 'Value of "this" must be of type %s', TypeError, HideStackFramesError);
 E('ERR_INVALID_TUPLE', '%s must be an iterable %s tuple', TypeError);
 E('ERR_INVALID_TYPESCRIPT_SYNTAX', '%s', SyntaxError);
 E('ERR_INVALID_URI', 'URI malformed', URIError);
@@ -1656,9 +1662,18 @@ E('ERR_QUIC_ENDPOINT_CLOSED', 'QUIC endpoint closed: %s (%d)', Error);
 E('ERR_QUIC_OPEN_STREAM_FAILED', 'Failed to open QUIC stream', Error);
 E('ERR_QUIC_TRANSPORT_ERROR', 'A QUIC transport error occurred. %d [%s]', Error);
 E('ERR_QUIC_VERSION_NEGOTIATION_ERROR', 'The QUIC session requires version negotiation', Error);
-E('ERR_REQUIRE_ASYNC_MODULE', 'require() cannot be used on an ESM ' +
+E('ERR_REQUIRE_ASYNC_MODULE', function(filename, parentFilename) {
+  let message = 'require() cannot be used on an ESM ' +
   'graph with top-level await. Use import() instead. To see where the' +
-  ' top-level await comes from, use --experimental-print-required-tla.', Error);
+  ' top-level await comes from, use --experimental-print-required-tla.';
+  if (parentFilename) {
+    message += `\n  From ${parentFilename} `;
+  }
+  if (filename) {
+    message += `\n  Requiring ${filename} `;
+  }
+  return message;
+}, Error);
 E('ERR_REQUIRE_CYCLE_MODULE', '%s', Error);
 E('ERR_REQUIRE_ESM',
   function(filename, hasEsmSyntax, parentPath = null, packageJsonPath = null) {
@@ -1739,21 +1754,6 @@ E('ERR_STREAM_WRAP', 'Stream has StringDecoder set or is in objectMode', Error);
 E('ERR_STREAM_WRITE_AFTER_END', 'write after end', Error);
 E('ERR_SYNTHETIC', 'JavaScript Callstack', Error);
 E('ERR_SYSTEM_ERROR', 'A system error occurred', SystemError, HideStackFramesError);
-E('ERR_TAP_LEXER_ERROR', function(errorMsg) {
-  hideInternalStackFrames(this);
-  return errorMsg;
-}, Error);
-E('ERR_TAP_PARSER_ERROR', function(errorMsg, details, tokenCausedError, source) {
-  hideInternalStackFrames(this);
-  this.cause = tokenCausedError;
-  const { column, line, start, end } = tokenCausedError.location;
-  const errorDetails = `${details} at line ${line}, column ${column} (start ${start}, end ${end})`;
-  return errorMsg + errorDetails;
-}, SyntaxError);
-E('ERR_TAP_VALIDATION_ERROR', function(errorMsg) {
-  hideInternalStackFrames(this);
-  return errorMsg;
-}, Error);
 E('ERR_TEST_FAILURE', function(error, failureType) {
   hideInternalStackFrames(this);
   assert(typeof failureType === 'string' || typeof failureType === 'symbol',
@@ -1809,6 +1809,8 @@ E('ERR_TRACE_EVENTS_CATEGORY_REQUIRED',
   'At least one category is required', TypeError);
 E('ERR_TRACE_EVENTS_UNAVAILABLE', 'Trace events are unavailable', Error);
 
+E('ERR_TRAILING_JUNK_AFTER_STREAM_END', 'Trailing junk found after the end of the compressed stream', TypeError);
+
 // This should probably be a `RangeError`.
 E('ERR_TTY_INIT_FAILED', 'TTY initialization failed', SystemError);
 E('ERR_UNAVAILABLE_DURING_EXIT', 'Cannot call function in process exit ' +
@@ -1853,6 +1855,7 @@ E('ERR_UNSUPPORTED_NODE_MODULES_TYPE_STRIPPING',
 E('ERR_UNSUPPORTED_RESOLVE_REQUEST',
   'Failed to resolve module specifier "%s" from "%s": Invalid relative URL or base scheme is not hierarchical.',
   TypeError);
+E('ERR_UNSUPPORTED_TYPESCRIPT_SYNTAX', '%s', SyntaxError);
 E('ERR_USE_AFTER_CLOSE', '%s was closed', Error);
 
 // This should probably be a `TypeError`.
@@ -1903,3 +1906,4 @@ E('ERR_WORKER_UNSERIALIZABLE_ERROR',
   'Serializing an uncaught exception failed', Error);
 E('ERR_WORKER_UNSUPPORTED_OPERATION',
   '%s is not supported in workers', TypeError);
+E('ERR_ZSTD_INVALID_PARAM', '%s is not a valid zstd parameter', RangeError);
