@@ -22,7 +22,15 @@ let debug = require('internal/util/debuglog').debuglog('esm', (fn) => {
   debug = fn;
 });
 
-const { ModuleWrap, kInstantiated } = internalBinding('module_wrap');
+const {
+  ModuleWrap,
+  kErrored,
+  kEvaluated,
+  kEvaluating,
+  kEvaluationPhase,
+  kInstantiated,
+  kUninstantiated,
+} = internalBinding('module_wrap');
 const {
   privateSymbols: {
     entry_point_module_private_symbol,
@@ -30,12 +38,13 @@ const {
 } = internalBinding('util');
 const { decorateErrorStack, kEmptyObject } = require('internal/util');
 const {
-  getSourceMapsEnabled,
+  getSourceMapsSupport,
 } = require('internal/source_map/source_map_cache');
 const assert = require('internal/assert');
 const resolvedPromise = PromiseResolve();
 const {
   setHasStartedUserESMExecution,
+  urlToFilename,
 } = require('internal/modules/helpers');
 const { getOptionValue } = require('internal/options');
 const noop = FunctionPrototype;
@@ -57,9 +66,49 @@ const isCommonJSGlobalLikeNotDefinedError = (errorMessage) =>
     (globalLike) => errorMessage === `${globalLike} is not defined`,
   );
 
+
+/**
+ *
+ * @param {Error} e
+ * @param {string} url
+ * @returns {void}
+ */
+const explainCommonJSGlobalLikeNotDefinedError = (e, url, hasTopLevelAwait) => {
+  if (e?.name === 'ReferenceError' &&
+      isCommonJSGlobalLikeNotDefinedError(e.message)) {
+
+    if (hasTopLevelAwait) {
+      e.message = `Cannot determine intended module format because both require() and top-level await are present. If the code is intended to be CommonJS, wrap await in an async function. If the code is intended to be an ES module, replace require() with import.`;
+      e.code = 'ERR_AMBIGUOUS_MODULE_SYNTAX';
+      return;
+    }
+
+    e.message += ' in ES module scope';
+
+    if (StringPrototypeStartsWith(e.message, 'require ')) {
+      e.message += ', you can use import instead';
+    }
+
+    const packageConfig =
+      StringPrototypeStartsWith(url, 'file://') &&
+        RegExpPrototypeExec(/\.js(\?[^#]*)?(#.*)?$/, url) !== null &&
+        require('internal/modules/package_json_reader')
+          .getPackageScopeConfig(url);
+    if (packageConfig.type === 'module') {
+      e.message +=
+        '\nThis file is being treated as an ES module because it has a ' +
+        `'.js' file extension and '${packageConfig.pjsonPath}' contains ` +
+        '"type": "module". To treat it as a CommonJS script, rename it ' +
+        'to use the \'.cjs\' file extension.';
+    }
+  }
+};
+
 class ModuleJobBase {
-  constructor(url, importAttributes, isMain, inspectBrk) {
+  constructor(url, importAttributes, phase, isMain, inspectBrk) {
+    assert(typeof phase === 'number');
     this.importAttributes = importAttributes;
+    this.phase = phase;
     this.isMain = isMain;
     this.inspectBrk = inspectBrk;
 
@@ -77,14 +126,15 @@ class ModuleJob extends ModuleJobBase {
    * @param {string} url URL of the module to be wrapped in ModuleJob.
    * @param {ImportAttributes} importAttributes Import attributes from the import statement.
    * @param {ModuleWrap|Promise<ModuleWrap>} moduleOrModulePromise Translated ModuleWrap for the module.
+   * @param {number} phase The phase to load the module to.
    * @param {boolean} isMain Whether the module is the entry point.
    * @param {boolean} inspectBrk Whether this module should be evaluated with the
    *                             first line paused in the debugger (because --inspect-brk is passed).
    * @param {boolean} isForRequireInImportedCJS Whether this is created for require() in imported CJS.
    */
-  constructor(loader, url, importAttributes = { __proto__: null },
-              moduleOrModulePromise, isMain, inspectBrk, isForRequireInImportedCJS = false) {
-    super(url, importAttributes, isMain, inspectBrk);
+  constructor(loader, url, importAttributes = { __proto__: null }, moduleOrModulePromise,
+              phase = kEvaluationPhase, isMain, inspectBrk, isForRequireInImportedCJS = false) {
+    super(url, importAttributes, phase, isMain, inspectBrk);
     this.#loader = loader;
 
     // Expose the promise to the ModuleWrap directly for linking below.
@@ -96,22 +146,37 @@ class ModuleJob extends ModuleJobBase {
       this.modulePromise = moduleOrModulePromise;
     }
 
-    // Promise for the list of all dependencyJobs.
-    this.linked = this._link();
-    // This promise is awaited later anyway, so silence
-    // 'unhandled rejection' warnings.
-    PromisePrototypeThen(this.linked, undefined, noop);
+    if (this.phase === kEvaluationPhase) {
+      // Promise for the list of all dependencyJobs.
+      this.linked = this.#link();
+      // This promise is awaited later anyway, so silence
+      // 'unhandled rejection' warnings.
+      PromisePrototypeThen(this.linked, undefined, noop);
+    }
 
     // instantiated == deep dependency jobs wrappers are instantiated,
     // and module wrapper is instantiated.
     this.instantiated = undefined;
   }
 
+  /**
+   * Ensure that this ModuleJob is moving towards the required phase
+   * (does not necessarily mean it is ready at that phase - run does that)
+   * @param {number} phase
+   */
+  ensurePhase(phase) {
+    if (this.phase < phase) {
+      this.phase = phase;
+      this.linked = this.#link();
+      PromisePrototypeThen(this.linked, undefined, noop);
+    }
+  }
+
   /**
    * Iterates the module requests and links with the loader.
    * @returns {Promise<ModuleJob[]>} Dependency module jobs.
    */
-  async _link() {
+  async #link() {
     this.module = await this.modulePromise;
     assert(this.module instanceof ModuleWrap);
 
@@ -122,43 +187,51 @@ class ModuleJob extends ModuleJobBase {
     // these `link` callbacks depending on each other.
     // Create an ArrayLike to avoid calling into userspace with `.then`
     // when returned from the async function.
-    const dependencyJobs = Array(moduleRequests.length);
-    ObjectSetPrototypeOf(dependencyJobs, null);
+    const evaluationDepJobs = Array(moduleRequests.length);
+    ObjectSetPrototypeOf(evaluationDepJobs, null);
 
-    // Specifiers should be aligned with the moduleRequests array in order.
-    const specifiers = Array(moduleRequests.length);
+    // Modules should be aligned with the moduleRequests array in order.
     const modulePromises = Array(moduleRequests.length);
+    // Track each loop for whether it is an evaluation phase or source phase request.
+    let isEvaluation;
     // Iterate with index to avoid calling into userspace with `Symbol.iterator`.
-    for (let idx = 0; idx < moduleRequests.length; idx++) {
-      const { specifier, attributes } = moduleRequests[idx];
+    for (
+      let idx = 0, eidx = 0;
+      // Use the let-scoped eidx value to update the executionDepJobs length at the end of the loop.
+      idx < moduleRequests.length || (evaluationDepJobs.length = eidx, false);
+      idx++, eidx += isEvaluation
+    ) {
+      const { specifier, phase, attributes } = moduleRequests[idx];
+      isEvaluation = phase === kEvaluationPhase;
       // TODO(joyeecheung): resolve all requests first, then load them in another
       // loop so that hooks can pre-fetch sources off-thread.
       const dependencyJobPromise = this.#loader.getModuleJobForImport(
-        specifier, this.url, attributes,
+        specifier, this.url, attributes, phase,
       );
       const modulePromise = PromisePrototypeThen(dependencyJobPromise, (job) => {
         debug(`async link() ${this.url} -> ${specifier}`, job);
-        dependencyJobs[idx] = job;
+        if (phase === kEvaluationPhase) {
+          evaluationDepJobs[eidx] = job;
+        }
         return job.modulePromise;
       });
       modulePromises[idx] = modulePromise;
-      specifiers[idx] = specifier;
     }
 
     const modules = await SafePromiseAllReturnArrayLike(modulePromises);
-    this.module.link(specifiers, modules);
+    this.module.link(modules);
 
-    return dependencyJobs;
+    return evaluationDepJobs;
   }
 
-  instantiate() {
+  #instantiate() {
     if (this.instantiated === undefined) {
-      this.instantiated = this._instantiate();
+      this.instantiated = this.#_instantiate();
     }
     return this.instantiated;
   }
 
-  async _instantiate() {
+  async #_instantiate() {
     const jobsInGraph = new SafeSet();
     const addJobsToDependencyGraph = async (moduleJob) => {
       debug(`async addJobsToDependencyGraph() ${this.url}`, moduleJob);
@@ -186,10 +259,10 @@ class ModuleJob extends ModuleJobBase {
       // of missing named export. This is currently not possible because
       // stack trace originates in module_job, not the file itself. A hidden
       // symbol with filename could be set in node_errors.cc to facilitate this.
-      if (!getSourceMapsEnabled() &&
+      if (!getSourceMapsSupport().enabled &&
           StringPrototypeIncludes(e.message,
                                   ' does not provide an export named')) {
-        const splitStack = StringPrototypeSplit(e.stack, '\n');
+        const splitStack = StringPrototypeSplit(e.stack, '\n', 2);
         const parentFileUrl = RegExpPrototypeSymbolReplace(
           /:\d+$/,
           splitStack[0],
@@ -245,23 +318,48 @@ class ModuleJob extends ModuleJobBase {
     }
   }
 
-  runSync() {
+  runSync(parent) {
+    assert(this.phase === kEvaluationPhase);
     assert(this.module instanceof ModuleWrap);
-    if (this.instantiated !== undefined) {
-      return { __proto__: null, module: this.module };
-    }
+    let status = this.module.getStatus();
 
-    this.module.instantiate();
-    this.instantiated = PromiseResolve();
-    const timeout = -1;
-    const breakOnSigint = false;
-    setHasStartedUserESMExecution();
-    this.module.evaluate(timeout, breakOnSigint);
-    return { __proto__: null, module: this.module };
+    debug('ModuleJob.runSync', this.module);
+    // FIXME(joyeecheung): this cannot fully handle < kInstantiated. Make the linking
+    // fully synchronous instead.
+    if (status === kUninstantiated) {
+      this.module.async = this.module.instantiateSync();
+      status = this.module.getStatus();
+    }
+    if (status === kInstantiated || status === kErrored) {
+      const filename = urlToFilename(this.url);
+      const parentFilename = urlToFilename(parent?.filename);
+      this.module.async ??= this.module.isGraphAsync();
+
+      if (this.module.async && !getOptionValue('--experimental-print-required-tla')) {
+        throw new ERR_REQUIRE_ASYNC_MODULE(filename, parentFilename);
+      }
+      if (status === kInstantiated) {
+        setHasStartedUserESMExecution();
+        const namespace = this.module.evaluateSync(filename, parentFilename);
+        return { __proto__: null, module: this.module, namespace };
+      }
+      throw this.module.getError();
+    } else if (status === kEvaluating || status === kEvaluated) {
+      // kEvaluating can show up when this is being used to deal with CJS <-> CJS cycles.
+      // Allow it for now, since we only need to ban ESM <-> CJS cycles which would be
+      // detected earlier during the linking phase, though the CJS handling in the ESM
+      // loader won't be able to emit warnings on pending circular exports like what
+      // the CJS loader does.
+      // TODO(joyeecheung): remove the re-invented require() in the ESM loader and
+      // always handle CJS using the CJS loader to eliminate the quirks.
+      return { __proto__: null, module: this.module, namespace: this.module.getNamespaceSync() };
+    }
+    assert.fail(`Unexpected module status ${status}.`);
   }
 
   async run(isEntryPoint = false) {
-    await this.instantiate();
+    assert(this.phase === kEvaluationPhase);
+    await this.#instantiate();
     if (isEntryPoint) {
       globalThis[entry_point_module_private_symbol] = this.module;
     }
@@ -271,27 +369,7 @@ class ModuleJob extends ModuleJobBase {
     try {
       await this.module.evaluate(timeout, breakOnSigint);
     } catch (e) {
-      if (e?.name === 'ReferenceError' &&
-          isCommonJSGlobalLikeNotDefinedError(e.message)) {
-        e.message += ' in ES module scope';
-
-        if (StringPrototypeStartsWith(e.message, 'require ')) {
-          e.message += ', you can use import instead';
-        }
-
-        const packageConfig =
-          StringPrototypeStartsWith(this.module.url, 'file://') &&
-            RegExpPrototypeExec(/\.js(\?[^#]*)?(#.*)?$/, this.module.url) !== null &&
-            require('internal/modules/package_json_reader')
-              .getPackageScopeConfig(this.module.url);
-        if (packageConfig.type === 'module') {
-          e.message +=
-            '\nThis file is being treated as an ES module because it has a ' +
-            `'.js' file extension and '${packageConfig.pjsonPath}' contains ` +
-            '"type": "module". To treat it as a CommonJS script, rename it ' +
-            'to use the \'.cjs\' file extension.';
-        }
-      }
+      explainCommonJSGlobalLikeNotDefinedError(e, this.module.url, this.module.hasTopLevelAwait());
       throw e;
     }
     return { __proto__: null, module: this.module };
@@ -316,40 +394,62 @@ class ModuleJobSync extends ModuleJobBase {
    * @param {string} url URL of the module to be wrapped in ModuleJob.
    * @param {ImportAttributes} importAttributes Import attributes from the import statement.
    * @param {ModuleWrap} moduleWrap Translated ModuleWrap for the module.
+   * @param {number} phase The phase to load the module to.
    * @param {boolean} isMain Whether the module is the entry point.
    * @param {boolean} inspectBrk Whether this module should be evaluated with the
    *                             first line paused in the debugger (because --inspect-brk is passed).
    */
-  constructor(loader, url, importAttributes, moduleWrap, isMain, inspectBrk) {
-    super(url, importAttributes, isMain, inspectBrk, true);
+  constructor(loader, url, importAttributes, moduleWrap, phase = kEvaluationPhase, isMain,
+              inspectBrk) {
+    super(url, importAttributes, phase, isMain, inspectBrk, true);
 
     this.#loader = loader;
     this.module = moduleWrap;
 
     assert(this.module instanceof ModuleWrap);
+    this.linked = undefined;
+    this.type = importAttributes.type;
+    if (phase === kEvaluationPhase) {
+      this.#link();
+    }
+  }
+
+  /**
+   * Ensure that this ModuleJob is at the required phase
+   * @param {number} phase
+   */
+  ensurePhase(phase) {
+    if (this.phase < phase) {
+      this.phase = phase;
+      this.#link();
+    }
+  }
+
+  #link() {
     // Store itself into the cache first before linking in case there are circular
     // references in the linking.
-    loader.loadCache.set(url, importAttributes.type, this);
-
+    this.#loader.loadCache.set(this.url, this.type, this);
     try {
       const moduleRequests = this.module.getModuleRequests();
-      // Specifiers should be aligned with the moduleRequests array in order.
-      const specifiers = Array(moduleRequests.length);
+      // Modules should be aligned with the moduleRequests array in order.
       const modules = Array(moduleRequests.length);
-      const jobs = Array(moduleRequests.length);
+      const evaluationDepJobs = Array(moduleRequests.length);
+      let j = 0;
       for (let i = 0; i < moduleRequests.length; ++i) {
-        const { specifier, attributes } = moduleRequests[i];
-        const job = this.#loader.getModuleJobForRequire(specifier, url, attributes);
-        specifiers[i] = specifier;
+        const { specifier, attributes, phase } = moduleRequests[i];
+        const job = this.#loader.getModuleJobForRequire(specifier, this.url, attributes, phase);
         modules[i] = job.module;
-        jobs[i] = job;
+        if (phase === kEvaluationPhase) {
+          evaluationDepJobs[j++] = job;
+        }
       }
-      this.module.link(specifiers, modules);
-      this.linked = jobs;
+      evaluationDepJobs.length = j;
+      this.module.link(modules);
+      this.linked = evaluationDepJobs;
     } finally {
       // Restore it - if it succeeds, we'll reset in the caller; Otherwise it's
       // not cached and if the error is caught, subsequent attempt would still fail.
-      loader.loadCache.delete(url, importAttributes.type);
+      this.#loader.loadCache.delete(this.url, this.type);
     }
   }
 
@@ -358,6 +458,7 @@ class ModuleJobSync extends ModuleJobBase {
   }
 
   async run() {
+    assert(this.phase === kEvaluationPhase);
     // This path is hit by a require'd module that is imported again.
     const status = this.module.getStatus();
     if (status > kInstantiated) {
@@ -380,7 +481,8 @@ class ModuleJobSync extends ModuleJobBase {
                 `Status = ${status}`);
   }
 
-  runSync() {
+  runSync(parent) {
+    assert(this.phase === kEvaluationPhase);
     // TODO(joyeecheung): add the error decoration logic from the async instantiate.
     this.module.async = this.module.instantiateSync();
     // If --experimental-print-required-tla is true, proceeds to evaluation even
@@ -389,12 +491,19 @@ class ModuleJobSync extends ModuleJobBase {
     // TODO(joyeecheung): track the asynchroniticy using v8::Module::HasTopLevelAwait()
     // and we'll be able to throw right after compilation of the modules, using acron
     // to find and print the TLA.
+    const parentFilename = urlToFilename(parent?.filename);
+    const filename = urlToFilename(this.url);
     if (this.module.async && !getOptionValue('--experimental-print-required-tla')) {
-      throw new ERR_REQUIRE_ASYNC_MODULE();
+      throw new ERR_REQUIRE_ASYNC_MODULE(filename, parentFilename);
     }
     setHasStartedUserESMExecution();
-    const namespace = this.module.evaluateSync();
-    return { __proto__: null, module: this.module, namespace };
+    try {
+      const namespace = this.module.evaluateSync(filename, parentFilename);
+      return { __proto__: null, module: this.module, namespace };
+    } catch (e) {
+      explainCommonJSGlobalLikeNotDefinedError(e, this.module.url, this.module.hasTopLevelAwait());
+      throw e;
+    }
   }
 }
 
