@@ -1,17 +1,16 @@
 'use strict';
 
 const {
-  ArrayPrototypeMap,
   ArrayPrototypePush,
-  Boolean,
   FunctionPrototypeCall,
   JSONParse,
-  ObjectKeys,
+  ObjectAssign,
   ObjectPrototypeHasOwnProperty,
   ReflectApply,
   SafeArrayIterator,
   SafeMap,
   SafeSet,
+  SafeWeakMap,
   StringPrototypeIncludes,
   StringPrototypeReplaceAll,
   StringPrototypeSlice,
@@ -52,6 +51,7 @@ let debug = require('internal/util/debuglog').debuglog('esm', (fn) => {
 });
 const { emitExperimentalWarning, kEmptyObject, setOwnProperty, isWindows } = require('internal/util');
 const {
+  ERR_INVALID_RETURN_PROPERTY_VALUE,
   ERR_UNKNOWN_BUILTIN_MODULE,
 } = require('internal/errors').codes;
 const { maybeCacheSourceMap } = require('internal/source_map/source_map_cache');
@@ -103,7 +103,8 @@ translators.set('module', function moduleStrategy(url, source, isMain) {
   source = stringify(source);
   debug(`Translating StandardModule ${url}`);
   const { compileSourceTextModule } = require('internal/modules/esm/utils');
-  const module = compileSourceTextModule(url, source, this);
+  const context = isMain ? { isMain } : undefined;
+  const module = compileSourceTextModule(url, source, this, context);
   return module;
 });
 
@@ -121,10 +122,10 @@ translators.set('module', function moduleStrategy(url, source, isMain) {
 function loadCJSModule(module, source, url, filename, isMain) {
   const compileResult = compileFunctionForCJSLoader(source, filename, false /* is_sea_main */, false);
 
-  const { function: compiledWrapper, sourceMapURL } = compileResult;
+  const { function: compiledWrapper, sourceMapURL, sourceURL } = compileResult;
   // Cache the source map for the cjs module if present.
   if (sourceMapURL) {
-    maybeCacheSourceMap(url, source, module, false, undefined, sourceMapURL);
+    maybeCacheSourceMap(url, source, module, false, sourceURL, sourceMapURL);
   }
 
   const cascadedLoader = require('internal/modules/esm/loader').getOrInitializeCascadedLoader();
@@ -185,13 +186,16 @@ function createCJSModuleWrap(url, source, isMain, format, loadCJS = loadCJSModul
   // In case the source was not provided by the `load` step, we need fetch it now.
   source = stringify(source ?? getSource(new URL(url)).source);
 
-  const { exportNames, module } = cjsPreparseModuleExports(filename, source, isMain, format);
+  const { exportNames, module } = cjsPreparseModuleExports(filename, source, format);
   cjsCache.set(url, module);
 
-  const wrapperNames = [...exportNames, 'module.exports'];
+  const wrapperNames = [...exportNames];
   if (!exportNames.has('default')) {
     ArrayPrototypePush(wrapperNames, 'default');
   }
+  if (!exportNames.has('module.exports')) {
+    ArrayPrototypePush(wrapperNames, 'module.exports');
+  }
 
   if (isMain) {
     setOwnProperty(process, 'mainModule', module);
@@ -212,7 +216,8 @@ function createCJSModuleWrap(url, source, isMain, format, loadCJS = loadCJSModul
       ({ exports } = module);
     }
     for (const exportName of exportNames) {
-      if (!ObjectPrototypeHasOwnProperty(exports, exportName) || exportName === 'default') {
+      if (exportName === 'default' || exportName === 'module.exports' ||
+          !ObjectPrototypeHasOwnProperty(exports, exportName)) {
         continue;
       }
       // We might trigger a getter -> dont fail.
@@ -229,6 +234,47 @@ function createCJSModuleWrap(url, source, isMain, format, loadCJS = loadCJSModul
   }, module);
 }
 
+/**
+ * Creates a ModuleWrap object for a CommonJS module without source texts.
+ * @param {string} url - The URL of the module.
+ * @param {boolean} isMain - Whether the module is the main module.
+ * @returns {ModuleWrap} The ModuleWrap object for the CommonJS module.
+ */
+function createCJSNoSourceModuleWrap(url, isMain) {
+  debug(`Translating CJSModule without source ${url}`);
+
+  const filename = urlToFilename(url);
+
+  const module = cjsEmplaceModuleCacheEntry(filename);
+  cjsCache.set(url, module);
+
+  if (isMain) {
+    setOwnProperty(process, 'mainModule', module);
+  }
+
+  // Addon export names are not known until the addon is loaded.
+  const exportNames = ['default', 'module.exports'];
+  return new ModuleWrap(url, undefined, exportNames, function evaluationCallback() {
+    debug(`Loading CJSModule ${url}`);
+
+    if (!module.loaded) {
+      wrapModuleLoad(filename, null, isMain);
+    }
+
+    /** @type {import('./loader').ModuleExports} */
+    let exports;
+    if (module[kModuleExport] !== undefined) {
+      exports = module[kModuleExport];
+      module[kModuleExport] = undefined;
+    } else {
+      ({ exports } = module);
+    }
+
+    this.setExport('default', exports);
+    this.setExport('module.exports', exports);
+  }, module);
+}
+
 translators.set('commonjs-sync', function requireCommonJS(url, source, isMain) {
   initCJSParseSync();
 
@@ -250,7 +296,6 @@ translators.set('require-commonjs', (url, source, isMain) => {
 // Handle CommonJS modules referenced by `require` calls.
 // This translator function must be sync, as `require` is sync.
 translators.set('require-commonjs-typescript', (url, source, isMain) => {
-  emitExperimentalWarning('Type Stripping');
   assert(cjsParse);
   const code = stripTypeScriptModuleTypes(stringify(source), url);
   return createCJSModuleWrap(url, code, isMain, 'commonjs-typescript');
@@ -280,26 +325,38 @@ translators.set('commonjs', function commonjsStrategy(url, source, isMain) {
   return createCJSModuleWrap(url, source, isMain, 'commonjs', cjsLoader);
 });
 
+/**
+ * Get or create an entry in the CJS module cache for the given filename.
+ * @param {string} filename CJS module filename
+ * @returns {CJSModule} the cached CJS module entry
+ */
+function cjsEmplaceModuleCacheEntry(filename, exportNames) {
+  // TODO: Do we want to keep hitting the user mutable CJS loader here?
+  let cjsMod = CJSModule._cache[filename];
+  if (cjsMod) {
+    return cjsMod;
+  }
+
+  cjsMod = new CJSModule(filename);
+  cjsMod.filename = filename;
+  cjsMod.paths = CJSModule._nodeModulePaths(cjsMod.path);
+  cjsMod[kIsCachedByESMLoader] = true;
+  CJSModule._cache[filename] = cjsMod;
+
+  return cjsMod;
+}
+
 /**
  * Pre-parses a CommonJS module's exports and re-exports.
  * @param {string} filename - The filename of the module.
  * @param {string} [source] - The source code of the module.
- * @param {boolean} isMain - Whether it is pre-parsing for the entry point.
- * @param {string} format
+ * @param {string} [format]
  */
-function cjsPreparseModuleExports(filename, source, isMain, format) {
-  let module = CJSModule._cache[filename];
-  if (module && module[kModuleExportNames] !== undefined) {
+function cjsPreparseModuleExports(filename, source, format) {
+  const module = cjsEmplaceModuleCacheEntry(filename);
+  if (module[kModuleExportNames] !== undefined) {
     return { module, exportNames: module[kModuleExportNames] };
   }
-  const loaded = Boolean(module);
-  if (!loaded) {
-    module = new CJSModule(filename);
-    module.filename = filename;
-    module.paths = CJSModule._nodeModulePaths(module.path);
-    module[kIsCachedByESMLoader] = true;
-    CJSModule._cache[filename] = module;
-  }
 
   if (source === undefined) {
     ({ source } = loadSourceForCJSWithHooks(module, filename, format));
@@ -340,7 +397,7 @@ function cjsPreparseModuleExports(filename, source, isMain, format) {
 
       if (format === 'commonjs' ||
         (!BuiltinModule.normalizeRequirableId(resolved) && findLongestRegisteredExtension(resolved) === '.js')) {
-        const { exportNames: reexportNames } = cjsPreparseModuleExports(resolved, undefined, false, format);
+        const { exportNames: reexportNames } = cjsPreparseModuleExports(resolved, undefined, format);
         for (const name of reexportNames) {
           exportNames.add(name);
         }
@@ -427,6 +484,14 @@ translators.set('json', function jsonStrategy(url, source) {
 });
 
 // Strategy for loading a wasm module
+// This logic should collapse into WebAssembly Module Record in future.
+/**
+ * @type {WeakMap<
+ *   import('internal/modules/esm/utils').ModuleNamespaceObject,
+ *   WebAssembly.Instance
+ * >} [[Instance]] slot proxy for WebAssembly Module Record
+ */
+const wasmInstances = new SafeWeakMap();
 translators.set('wasm', async function(url, source) {
   emitExperimentalWarning('Importing WebAssembly modules');
 
@@ -445,37 +510,100 @@ translators.set('wasm', async function(url, source) {
     throw err;
   }
 
-  const imports =
-      ArrayPrototypeMap(WebAssembly.Module.imports(compiled),
-                        ({ module }) => module);
-  const exports =
-    ArrayPrototypeMap(WebAssembly.Module.exports(compiled),
-                      ({ name }) => name);
-
-  const createDynamicModule = require(
-    'internal/modules/esm/create_dynamic_module');
-  return createDynamicModule(imports, exports, url, (reflect) => {
-    const { exports } = new WebAssembly.Instance(compiled, reflect.imports);
-    for (const expt of ObjectKeys(exports)) {
-      reflect.exports[expt].set(exports[expt]);
+  const importsList = new SafeSet();
+  const wasmGlobalImports = [];
+  for (const impt of WebAssembly.Module.imports(compiled)) {
+    if (impt.kind === 'global') {
+      ArrayPrototypePush(wasmGlobalImports, impt);
     }
-  }).module;
+    importsList.add(impt.module);
+  }
+
+  const exportsList = new SafeSet();
+  const wasmGlobalExports = new SafeSet();
+  for (const expt of WebAssembly.Module.exports(compiled)) {
+    if (expt.kind === 'global') {
+      wasmGlobalExports.add(expt.name);
+    }
+    exportsList.add(expt.name);
+  }
+
+  const createDynamicModule = require('internal/modules/esm/create_dynamic_module');
+
+  const { module } = createDynamicModule([...importsList], [...exportsList], url, (reflect) => {
+    for (const impt of importsList) {
+      const importNs = reflect.imports[impt];
+      const wasmInstance = wasmInstances.get(importNs);
+      if (wasmInstance) {
+        const wrappedModule = ObjectAssign({ __proto__: null }, reflect.imports[impt]);
+        for (const { module, name } of wasmGlobalImports) {
+          if (module !== impt) {
+            continue;
+          }
+          // Import of Wasm module global -> get direct WebAssembly.Global wrapped value.
+          // JS API validations otherwise remain the same.
+          wrappedModule[name] = wasmInstance[name];
+        }
+        reflect.imports[impt] = wrappedModule;
+      }
+    }
+    // In cycles importing unexecuted Wasm, wasmInstance will be undefined, which will fail during
+    // instantiation, since all bindings will be in the Temporal Deadzone (TDZ).
+    const { exports } = new WebAssembly.Instance(compiled, reflect.imports);
+    wasmInstances.set(module.getNamespace(), exports);
+    for (const expt of exportsList) {
+      let val = exports[expt];
+      // Unwrap WebAssembly.Global for JS bindings
+      if (wasmGlobalExports.has(expt)) {
+        try {
+          // v128 will throw in GetGlobalValue, see:
+          // https://webassembly.github.io/esm-integration/js-api/index.html#getglobalvalue
+          val = val.value;
+        } catch {
+          // v128 doesn't support ToJsValue() -> use undefined (ideally should stay in TDZ)
+          continue;
+        }
+      }
+      reflect.exports[expt].set(val);
+    }
+  });
+  // WebAssembly modules support source phase imports, to import the compiled module
+  // separate from the linked instance.
+  module.setModuleSourceObject(compiled);
+  return module;
+});
+
+// Strategy for loading a addon
+translators.set('addon', function translateAddon(url, source, isMain) {
+  emitExperimentalWarning('Importing addons');
+
+  // The addon must be loaded from file system with dlopen. Assert
+  // the source is null.
+  if (source !== null) {
+    throw new ERR_INVALID_RETURN_PROPERTY_VALUE(
+      'null',
+      'load',
+      'source',
+      source);
+  }
+
+  debug(`Translating addon ${url}`);
+
+  return createCJSNoSourceModuleWrap(url, isMain);
 });
 
 // Strategy for loading a commonjs TypeScript module
-translators.set('commonjs-typescript', function(url, source) {
-  emitExperimentalWarning('Type Stripping');
+translators.set('commonjs-typescript', function(url, source, isMain) {
   assertBufferSource(source, true, 'load');
   const code = stripTypeScriptModuleTypes(stringify(source), url);
   debug(`Translating TypeScript ${url}`);
-  return FunctionPrototypeCall(translators.get('commonjs'), this, url, code, false);
+  return FunctionPrototypeCall(translators.get('commonjs'), this, url, code, isMain);
 });
 
 // Strategy for loading an esm TypeScript module
-translators.set('module-typescript', function(url, source) {
-  emitExperimentalWarning('Type Stripping');
+translators.set('module-typescript', function(url, source, isMain) {
   assertBufferSource(source, true, 'load');
   const code = stripTypeScriptModuleTypes(stringify(source), url);
   debug(`Translating TypeScript ${url}`);
-  return FunctionPrototypeCall(translators.get('module'), this, url, code, false);
+  return FunctionPrototypeCall(translators.get('module'), this, url, code, isMain);
 });
