@@ -10,12 +10,10 @@ const {
   RegExpPrototypeExec,
   SafeSet,
   SafeStringIterator,
-  StringPrototypeEndsWith,
   StringPrototypeIndexOf,
   StringPrototypeLastIndexOf,
   StringPrototypeReplaceAll,
   StringPrototypeSlice,
-  StringPrototypeStartsWith,
   StringPrototypeToLowerCase,
   StringPrototypeTrim,
   Symbol,
@@ -37,6 +35,11 @@ const {
   moveCursor,
 } = require('internal/readline/callbacks');
 
+const {
+  kIsMultiline,
+  kSetLine,
+} = require('internal/readline/interface');
+
 const {
   commonPrefix,
   kSubstringSearch,
@@ -298,8 +301,7 @@ function setupPreview(repl, contextSymbol, bufferSymbol, active) {
   function getInputPreview(input, callback) {
     // For similar reasons as `defaultEval`, wrap expressions starting with a
     // curly brace with parenthesis.
-    if (StringPrototypeStartsWith(input, '{') &&
-        !StringPrototypeEndsWith(input, ';') && !wrapped) {
+    if (!wrapped && input[0] === '{' && input[input.length - 1] !== ';') {
       input = `(${input})`;
       wrapped = true;
     }
@@ -364,8 +366,11 @@ function setupPreview(repl, contextSymbol, bufferSymbol, active) {
   }
 
   const showPreview = (showCompletion = true) => {
-    // Prevent duplicated previews after a refresh.
-    if (inputPreview !== null || !repl.isCompletionEnabled || !process.features.inspector) {
+    // Prevent duplicated previews after a refresh or in a multiline command.
+    if (inputPreview !== null ||
+        repl[kIsMultiline] ||
+        !repl.isCompletionEnabled ||
+        !process.features.inspector) {
       return;
     }
 
@@ -436,9 +441,9 @@ function setupPreview(repl, contextSymbol, bufferSymbol, active) {
 
       // Line breaks are very rare and probably only occur in case of error
       // messages with line breaks.
-      const lineBreakPos = StringPrototypeIndexOf(inspected, '\n');
-      if (lineBreakPos !== -1) {
-        inspected = `${StringPrototypeSlice(inspected, 0, lineBreakPos)}`;
+      const lineBreakMatch = RegExpPrototypeExec(/[\r\n\v]/, inspected);
+      if (lineBreakMatch !== null) {
+        inspected = `${StringPrototypeSlice(inspected, 0, lineBreakMatch.index)}`;
       }
 
       const result = repl.useColors ?
@@ -447,9 +452,14 @@ function setupPreview(repl, contextSymbol, bufferSymbol, active) {
 
       const { cursorPos, displayPos } = getPreviewPos();
       const rows = displayPos.rows - cursorPos.rows;
+      // Moves one line below all the user lines
       moveCursor(repl.output, 0, rows);
+      // Writes the preview there
       repl.output.write(`\n${result}`);
+
+      // Go back to the horizontal position of the cursor
       cursorTo(repl.output, cursorPos.cols);
+      // Go back to the vertical position of the cursor
       moveCursor(repl.output, 0, -rows - 1);
     };
 
@@ -675,7 +685,7 @@ function setupReverseSearch(repl) {
     // line to the found entry.
     if (!isInReverseSearch) {
       if (lastMatch !== -1) {
-        repl.line = repl.history[lastMatch];
+        repl[kSetLine](repl.history[lastMatch]);
         repl.cursor = lastCursor;
         repl.historyIndex = lastMatch;
       }
@@ -742,6 +752,22 @@ function setupReverseSearch(repl) {
   return { reverseSearch };
 }
 
+const startsWithBraceRegExp = /^\s*{/;
+const endsWithSemicolonRegExp = /;\s*$/;
+
+/**
+ * Checks if some provided code represents an object literal.
+ * This is helpful to prevent confusing repl code evaluations where
+ * strings such as `{ a : 1 }` would get interpreted as block statements
+ * rather than object literals.
+ * @param {string} code the code to check
+ * @returns {boolean} true if the code represents an object literal, false otherwise
+ */
+function isObjectLiteral(code) {
+  return RegExpPrototypeExec(startsWithBraceRegExp, code) !== null &&
+    RegExpPrototypeExec(endsWithSemicolonRegExp, code) === null;
+}
+
 module.exports = {
   REPL_MODE_SLOPPY: Symbol('repl-sloppy'),
   REPL_MODE_STRICT,
@@ -749,4 +775,5 @@ module.exports = {
   kStandaloneREPL: Symbol('kStandaloneREPL'),
   setupPreview,
   setupReverseSearch,
+  isObjectLiteral,
 };
