@@ -2,6 +2,8 @@
 const {
   ArrayPrototypePush,
   ArrayPrototypeSlice,
+  AtomicsStore,
+  AtomicsWait,
   Error,
   FunctionPrototypeBind,
   FunctionPrototypeCall,
@@ -18,10 +20,6 @@ const {
   StringPrototypeSlice,
   StringPrototypeStartsWith,
   globalThis: {
-    Atomics: {
-      store: AtomicsStore,
-      wait: AtomicsWait,
-    },
     SharedArrayBuffer,
   },
 } = primordials;
@@ -70,7 +68,14 @@ const kMockUnknownMessage = 3;
 const kWaitTimeout = 5_000;
 const kBadExportsMessage = 'Cannot create mock because named exports ' +
   'cannot be applied to the provided default export.';
-const kSupportedFormats = ['builtin', 'commonjs', 'module', 'module-typescript', 'commonjs-typescript'];
+const kSupportedFormats = [
+  'builtin',
+  'commonjs-typescript',
+  'commonjs',
+  'json',
+  'module-typescript',
+  'module',
+];
 let sharedModuleState;
 
 class MockFunctionContext {
@@ -277,6 +282,134 @@ class MockModuleContext {
 
 const { restore: restoreModule } = MockModuleContext.prototype;
 
+class MockPropertyContext {
+  #object;
+  #propertyName;
+  #value;
+  #originalValue;
+  #descriptor;
+  #accesses;
+  #onceValues;
+
+  constructor(object, propertyName, value) {
+    this.#onceValues = new SafeMap();
+    this.#accesses = [];
+    this.#object = object;
+    this.#propertyName = propertyName;
+    this.#originalValue = object[propertyName];
+    this.#value = arguments.length > 2 ? value : this.#originalValue;
+    this.#descriptor = ObjectGetOwnPropertyDescriptor(object, propertyName);
+    if (!this.#descriptor) {
+      throw new ERR_INVALID_ARG_VALUE(
+        'propertyName', propertyName, 'is not a property of the object',
+      );
+    }
+
+    const { configurable, enumerable } = this.#descriptor;
+    ObjectDefineProperty(object, propertyName, {
+      __proto__: null,
+      configurable,
+      enumerable,
+      get: () => {
+        const nextValue = this.#getAccessValue(this.#value);
+        const access = {
+          __proto__: null,
+          type: 'get',
+          value: nextValue,
+          // eslint-disable-next-line no-restricted-syntax
+          stack: new Error(),
+        };
+        ArrayPrototypePush(this.#accesses, access);
+        return nextValue;
+      },
+      set: this.mockImplementation.bind(this),
+    });
+  }
+
+  /**
+   * Gets an array of recorded accesses (get/set) to the property.
+   * @returns {Array} An array of access records.
+   */
+  get accesses() {
+    return ArrayPrototypeSlice(this.#accesses, 0);
+  }
+
+  /**
+   * Retrieves the number of times the property was accessed (get or set).
+   * @returns {number} The total number of accesses.
+   */
+  accessCount() {
+    return this.#accesses.length;
+  }
+
+  /**
+   * Sets a new value for the property.
+   * @param {any} value - The new value to be set.
+   * @throws {Error} If the property is not writable.
+   */
+  mockImplementation(value) {
+    if (!this.#descriptor.writable) {
+      throw new ERR_INVALID_ARG_VALUE(
+        'propertyName', this.#propertyName, 'cannot be set',
+      );
+    }
+    const nextValue = this.#getAccessValue(value);
+    const access = {
+      __proto__: null,
+      type: 'set',
+      value: nextValue,
+      // eslint-disable-next-line no-restricted-syntax
+      stack: new Error(),
+    };
+    ArrayPrototypePush(this.#accesses, access);
+    this.#value = nextValue;
+  }
+
+  #getAccessValue(value) {
+    const accessIndex = this.#accesses.length;
+    let accessValue;
+    if (this.#onceValues.has(accessIndex)) {
+      accessValue = this.#onceValues.get(accessIndex);
+      this.#onceValues.delete(accessIndex);
+    } else {
+      accessValue = value;
+    }
+    return accessValue;
+  }
+
+  /**
+   * Sets a value to be used only for the next access (get or set), or a specific access index.
+   * @param {any} value - The value to be used once.
+   * @param {number} [onAccess] - The access index to be replaced.
+   */
+  mockImplementationOnce(value, onAccess) {
+    const nextAccess = this.#accesses.length;
+    const accessIndex = onAccess ?? nextAccess;
+    validateInteger(accessIndex, 'onAccess', nextAccess);
+    this.#onceValues.set(accessIndex, value);
+  }
+
+  /**
+   * Resets the recorded accesses to the property.
+   */
+  resetAccesses() {
+    this.#accesses = [];
+  }
+
+  /**
+   * Restores the original value of the property that was mocked.
+   */
+  restore() {
+    ObjectDefineProperty(this.#object, this.#propertyName, {
+      __proto__: null,
+      ...this.#descriptor,
+      value: this.#originalValue,
+    });
+  }
+}
+
+const { restore: restoreProperty } = MockPropertyContext.prototype;
+
 class MockTracker {
   #mocks = [];
   #timers;
@@ -538,7 +671,7 @@ class MockTracker {
 
     if (baseURL.searchParams.has(kMockSearchParam)) {
       throw new ERR_INVALID_STATE(
-        `Cannot mock '${specifier}.' The module is already mocked.`,
+        `Cannot mock '${specifier}'. The module is already mocked.`,
       );
     }
 
@@ -566,6 +699,41 @@ class MockTracker {
     return ctx;
   }
 
+  /**
+   * Creates a property tracker for a specified object.
+   * @param {(object)} object - The object whose value is being tracked.
+   * @param {string} propertyName - The identifier of the property on object to be tracked.
+   * @param {any} value - An optional replacement value used as the mock value for object[valueName].
+   * @returns {ProxyConstructor} The mock property tracker.
+   */
+  property(
+    object,
+    propertyName,
+    value,
+  ) {
+    validateObject(object, 'object');
+    validateStringOrSymbol(propertyName, 'propertyName');
+
+    const ctx = arguments.length > 2 ?
+      new MockPropertyContext(object, propertyName, value) :
+      new MockPropertyContext(object, propertyName);
+    ArrayPrototypePush(this.#mocks, {
+      __proto__: null,
+      ctx,
+      restore: restoreProperty,
+    });
+
+    return new Proxy(object, {
+      __proto__: null,
+      get(target, property, receiver) {
+        if (property === 'mock') {
+          return ctx;
+        }
+        return ReflectGet(target, property, receiver);
+      },
+    });
+  }
+
   /**
    * Resets the mock tracker, restoring all mocks and clearing timers.
    */
