@@ -87,6 +87,8 @@ const {
 } = require('internal/test_runner/utils');
 const { Glob } = require('internal/fs/glob');
 const { once } = require('events');
+const { validatePath } = require('internal/fs/utils');
+const { loadPreloadModules } = require('internal/process/pre_execution');
 const {
   triggerUncaughtException,
   exitCodes: { kGenericUserError },
@@ -96,8 +98,17 @@ let debug = require('internal/util/debuglog').debuglog('test_runner', (fn) => {
 });
 
 const kIsolatedProcessName = Symbol('kIsolatedProcessName');
-const kFilterArgs = ['--test', '--experimental-test-coverage', '--watch'];
-const kFilterArgValues = ['--test-reporter', '--test-reporter-destination'];
+const kFilterArgs = [
+  '--test',
+  '--experimental-test-coverage',
+  '--watch',
+  '--experimental-default-config-file',
+];
+const kFilterArgValues = [
+  '--test-reporter',
+  '--test-reporter-destination',
+  '--experimental-config-file',
+];
 const kDiagnosticsFilterArgs = ['tests', 'suites', 'pass', 'fail', 'cancelled', 'skipped', 'todo', 'duration_ms'];
 
 const kCanceledTests = new SafeSet()
@@ -137,6 +148,7 @@ function getRunArgs(path, { forceExit,
                             only,
                             argv: suppliedArgs,
                             execArgv,
+                            root: { timeout },
                             cwd }) {
   const argv = ArrayPrototypeFilter(process.execArgv, filterExecArgv);
   if (forceExit === true) {
@@ -154,6 +166,9 @@ function getRunArgs(path, { forceExit,
   if (only === true) {
     ArrayPrototypePush(argv, '--test-only');
   }
+  if (timeout != null) {
+    ArrayPrototypePush(argv, `--test-timeout=${timeout}`);
+  }
 
   ArrayPrototypePushApply(argv, execArgv);
 
@@ -191,6 +206,7 @@ class FileTest extends Test {
       column: 1,
       file: resolve(this.name),
     };
+    this.timeout = null;
   }
 
   #skipReporting() {
@@ -478,6 +494,7 @@ function watchFiles(testFiles, opts) {
       // Reset the topLevel counter
       opts.root.harness.counters.topLevel = 0;
     }
+
     await runningSubtests.get(file);
     runningSubtests.set(file, runTestFile(file, filesWatcher, opts));
   }
@@ -502,7 +519,12 @@ function watchFiles(testFiles, opts) {
         return; // Avoid rerunning files when file deleted
       }
     }
+    // Reset the root start time to recalculate the duration
+    // of the run
+    opts.root.clearExecutionTime();
+    opts.root.reporter[kEmitMessage]('test:watch:restarted');
 
+    // Restart test files
     if (opts.isolation === 'none') {
       PromisePrototypeThen(restartTestFile(kIsolatedProcessName), undefined, (error) => {
         triggerUncaughtException(error, true /* fromPromise */);
@@ -555,6 +577,7 @@ function run(options = kEmptyObject) {
     isolation = 'process',
     watch,
     setup,
+    globalSetupPath,
     only,
     globPatterns,
     coverage = false,
@@ -664,6 +687,10 @@ function run(options = kEmptyObject) {
   validateStringArray(argv, 'options.argv');
   validateStringArray(execArgv, 'options.execArgv');
 
+  if (globalSetupPath != null) {
+    validatePath(globalSetupPath, 'options.globalSetupPath');
+  }
+
   const rootTestOptions = { __proto__: null, concurrency, timeout, signal };
   const globalOptions = {
     __proto__: null,
@@ -678,9 +705,11 @@ function run(options = kEmptyObject) {
     branchCoverage: branchCoverage,
     functionCoverage: functionCoverage,
     cwd,
+    globalSetupPath,
   };
   const root = createTestTree(rootTestOptions, globalOptions);
   let testFiles = files ?? createTestFileList(globPatterns, cwd);
+  const { isTestRunner } = globalOptions;
 
   if (shard) {
     testFiles = ArrayPrototypeFilter(testFiles, (_, index) => index % shard.total === shard.index - 1);
@@ -750,8 +779,20 @@ function run(options = kEmptyObject) {
           const cascadedLoader = esmLoader.getOrInitializeCascadedLoader();
           let topLevelTestCount = 0;
 
-          root.harness.bootstrapPromise = promise;
+          root.harness.bootstrapPromise = root.harness.bootstrapPromise ?
+            SafePromiseAllReturnVoid([root.harness.bootstrapPromise, promise]) :
+            promise;
 
+          // We need to setup the user modules in the test runner if we are running with
+          // --test-isolation=none and --test in order to avoid loading the user modules
+          // BEFORE the creation of the root test (that would cause them to get lost).
+          if (isTestRunner) {
+            // If we are not coming from the test runner entry point, the user-required and imported
+            // modules have already been loaded.
+            // Since it's possible to delete modules from require.cache, a CommonJS module
+            // could otherwise be executed twice.
+            loadPreloadModules();
+          }
           const userImports = getOptionValue('--import');
           for (let i = 0; i < userImports.length; i++) {
             await cascadedLoader.import(userImports[i], parentURL, kEmptyObject);
@@ -795,12 +836,15 @@ function run(options = kEmptyObject) {
         debug('beginning test execution');
         root.entryFile = null;
         finishBootstrap();
-        root.processPendingSubtests();
+        return root.processPendingSubtests();
       };
     }
   }
 
   const runChain = async () => {
+    if (root.harness?.bootstrapPromise) {
+      await root.harness.bootstrapPromise;
+    }
     if (typeof setup === 'function') {
       await setup(root.reporter);
     }
